// $Id: JareEngine.java,v 1.12 2001/08/12 18:21:24 ioerger Exp $

package Jare;
import java.util.Hashtable;
import java.util.Stack;
import java.util.Vector;

public class JareEngine
{
  Vector KB;
  Stack choices; // as unwind, push choice points (KB indexes); use for next()
  Stack saved; // save initial goal stack for next()
  int varCnt; // for creating unique new variable names
  boolean trace=false; // if true, print stack and env for each call to query1
  Hashtable hash;
  JareEnv lastEnv; // if passed into query(), save for calls to next()

  public JareEngine()
  {
    varCnt = 1; 
    saved = null; // to prevent calling next() before query()
    KB = new Vector();
    hash = new Hashtable(); // predicate -> list of clause indexes
  }

  public Hashtable get_hash() { return hash; }

  public void setTrace(boolean x) { trace = x; }

  public void readKBfile(String filename)
  {
    String s=J.read_file(filename);
    if (s.length()==0) warn("file not found");
    else 
    {
      System.out.println("reading: "+filename);
      assert(s);
    }
  }

  public void reset()
  {
    KB = new Vector();
    // forget about choices, saved, and lastEnv
    // leave settings for varCnt and trace
    hash = new Hashtable();
  }

  public void assert(String str)
  {
    try { addClauses(J.parse(str)); }
    catch (JareParseError p) { System.out.println(p); }
  }

  // retract any clauses whose head unifies

  public void retract(Vector fact) 
  {
    int n=KB.size();
    // could go through indexes for this predicate name, not unify with all 
    for (int i=n-1 ; i>=0 ; i--)
    {
      Vector clause=(Vector)(KB.elementAt(i));
      Vector head=(Vector)(clause.elementAt(0));
      if ((new JareEnv()).unify(fact,head)!=null) KB.removeElementAt(i);
    }
    rehash(); // could be inefficient if there are lots of retracts!
  }

  public void addFactStr(String str)
  {
    try { addClauseAndHash(J.parse(str)); } // parse adds extra set of parens
    catch (JareParseError p) { System.out.println(p); }
  }

  public void addFact(Vector fact)  
  {
    Vector v=new Vector(); 
    v.addElement(fact);
    addClauseAndHash(v);
  }

  public void rehash()
  {
    Vector KB2=KB;
    KB = new Vector();
    hash = new Hashtable();
    addClauses(KB2);
  } 

  public void addClauses(Vector clauses)
  {
    validateSyntax(clauses);
    for (int i=0 ; i<clauses.size() ; i++)
      addClauseAndHash((Vector)(clauses.elementAt(i)));
  }

  public void addClauseAndHash(Vector clause)
  {
    int i=KB.size();
    KB.add(clause);
    Vector head=(Vector)(clause.elementAt(0));
    String pred=(String)head.elementAt(0);
    Vector idxs=(Vector)hash.get(pred);
    if (idxs==null) idxs = new Vector();
    idxs.addElement(new Integer(i));
    hash.put(pred,idxs);
  }

  void validateSyntax(Vector KB)
  {
    for (int i=0 ; i<KB.size() ; i++)
    {
      if (!J.isVector(KB.elementAt(i)))
        die("KB must be set of non-empty Horn clauses"); 
      Vector clause=(Vector)KB.elementAt(i);
      if ((clause.size()==0) || (!J.isVector(clause.elementAt(0))))
        die("KB must be set of non-empty Horn clauses");
      Vector head=(Vector)clause.elementAt(0);
      if ((head.size()==0) || (!J.isString(head.elementAt(0))))
        die("KB must be set of non-empty Horn clauses");
    }
  }

  public void printKB()
  {
    for (int i=0 ; i<KB.size() ; i++)
      System.out.println(J.treeToString(KB.elementAt(i)));
  }

  Stack copyStack(Stack s1)
  {
    Stack s2=new Stack();
    for (int i=0 ; i<s1.size() ; i++)
      s2.push(s1.elementAt(i));
    return s2;
  }

  public JareEnv next()
  {
    // don't reset choices!
    if (saved==null) return null;
    Stack stack=copyStack(saved);
    JareEnv e2=new JareEnv(lastEnv); // saved from query, copy
    return query1(stack,e2); 
  }

  Vector standardizeVariablesApart(Vector clause)
  {
    JareEnv env=new JareEnv();
    return (Vector)sva(clause,env); // assume original clause is a Vector
  }

  // standardize variables apart each time a rule is used: e.g. ??V:1592
  // object is parse tree
  // for each unique var, create new var name and replace throughout
  // non-destructive: build new copy recursively

  Object sva(Object o,JareEnv env)
  {
    if (!J.isVector(o))
    {
      String s=(String)o;
      if (!J.isVar(s)) return s;
      String t=(String)env.lookup(s); // I know I only put strings in here
      if (t==null) { t = newVar(); env.bind(s,t); }
      return t;
    }
    else
    {
      Vector v1=(Vector)o;
      Vector v2=new Vector();
      for (int i=0 ; i<v1.size() ; i++)
      v2.addElement(sva(v1.elementAt(i),env));
      return v2;
    }
  }

  String newVar()
  {
    return "??V:"+varCnt++;
  }

  //--------------------------------------

  // tree represents parsed version of initial goal stack

  public JareEnv query(Vector tree)
  {
    return queryWithEnv(tree,null);
  }

  public JareEnv queryWithEnv(Vector tree,JareEnv e1)
  {
    if (e1==null) lastEnv = new JareEnv();
    else lastEnv = e1; // save
    JareEnv e2=new JareEnv(lastEnv); // copy
    choices = null;
    Stack stack=new Stack();
    saved = new Stack();
    for (int i=0 ; i<tree.size() ; i++) 
      stack.push(tree.elementAt(tree.size()-i-1)); // push in reverse order
    saved = copyStack(stack);
    return query1(stack,e2);
  }

  // special case, don't update saved goal stack
  // reset choice points for this search, but don't alter them externally
  
  JareEnv negation_search(Stack stack,JareEnv env)
  {
    Stack ch=choices;
    choices = null;
    JareEnv e=query1(stack,env);
    choices = ch;
    return e;
  }

  JareEnv query1(Stack stack,JareEnv env)
  {
    if (trace==true)
    {
      System.out.println("stack="+stack); // note: not substituted
      System.out.println("env="+env.hash);
      System.out.println();
    }

    int start=0;
    if (choices!=null) 
    {
      // fail from last step in proof for previous solution
      if (stack.empty()) 
      { 
        choices = null; 
        return null; 
      } 

      start = ((Integer)choices.pop()).intValue();
      if (start==-1) //(start==-1) // -1=negation, -2=procedural attachment 
      {
        Object neg=stack.pop();
        return unwind(query1(stack,env),stack,neg,-1);
      }
    }

    if (stack.empty()) 
    { 
      choices = new Stack(); 
      return env; 
    }

    Object o=stack.pop();
    if (!J.isVector(o)) return unwind_fail(stack,o);
    Vector goal=(Vector)o;
    String predName=(String)goal.elementAt(0);

    // negation: (assume not re-tracing for next() at this point)

    if (J.stringEq((String)goal.elementAt(0),"not"))
    {
      Stack temp=new Stack();
      temp.push(goal.elementAt(1));
      if (negation_search(temp,new JareEnv(env))==null) // copy env
        return unwind(query1(stack,env),stack,goal,-1); // stack already popped
      else return unwind_fail(stack,goal);
    }

    if (isProc(predName)) 
    {
      JareEnv e4=procAtt(goal,env);
      if (e4==null) return unwind_fail(stack,goal);
      else return unwind(query1(stack,e4),stack,goal,-2);
    }

    Vector ruleIdxs=(Vector)hash.get(predName);
    if (ruleIdxs==null) ruleIdxs = new Vector(); // just so loop doesn't fail
    for (int i=start ; i<ruleIdxs.size() ; i++)
    {
      Integer idx=(Integer)ruleIdxs.elementAt(i);
      Vector clause=(Vector)KB.elementAt(idx.intValue());
      String predicate=(String)((Vector)clause.elementAt(0)).elementAt(0);
      // if (!J.stringEq(predName,predicate)) continue; // no longer needed!
      clause = standardizeVariablesApart(clause);
      Vector head=(Vector)clause.elementAt(0);

      JareEnv e2=env.unify(goal,head);
      if (e2!=null)
      {
        int n=clause.size();
        for (int j=0 ; j<n-1 ; j++)
	    stack.push(clause.elementAt(n-j-1)); // crucial to push in reverse!
        JareEnv e3=query1(stack,e2);
        if (e3!=null) return unwind_success(e3,i);
        for (int j=0 ; j<n-1 ; j++) stack.pop();
      }
    }
    return unwind_fail(stack,goal);
  }

  //--------------------------------------

  JareEnv unwind(JareEnv env,Stack stack,Object goal,int cp)
  {
    if (env==null) return unwind_fail(stack,goal);
    else return unwind_success(env,cp);
  }

  JareEnv unwind_success(JareEnv env,int cp)
  {
    choices.push(new Integer(cp));
    return env;
  }

  JareEnv unwind_fail(Stack stack,Object goal)
  {
    stack.push(goal);
    return null;
  }

  //-------------------------------------------------
  // PROCEDURAL ATTACHMENT
  //-------------------------------------------------

  // generally speaking, procedural attachments:
  //   will return die with the wrong number of args (not just fail)
  //   will check truth if all args are instantiated
  //   will compute variable bindings for certain uninstantiated args

  boolean isProc(String s)
  {
    return (isMathop(s) || isListOp(s) || isMathEqPred(s)); // skip isType()
  }

  // is Eq is mathEqPred, must come before call to procAttMathEq()
  // is Cons is ListOp, must come before call to procAttList()

  JareEnv procAtt(Vector goal,JareEnv env)
  {
    String predName=(String)goal.elementAt(0);
    if (isMathop(predName)) return procAttMath(goal,env);
    else if (J.stringEq(predName,"eq")) return procAttStrEq(goal,env);
    else if (isMathEqPred(predName)) return procAttMathEq(goal,env);
    else if (J.stringEq(predName,"cons")) return procAttCons(goal,env);
    else if (isListOp(predName)) return procAttList(goal,env);
    else return null;
  }

  // = is for numbers; eq is for atoms
  // both args must be instantiated; lists and structures can't be tested
  
  boolean isMathEqPred(String s)
  {
    if (J.stringEq(s,"eq")) return true;
    if (J.stringEq(s,"=")) return true;
    if (J.stringEq(s,"<")) return true;
    if (J.stringEq(s,">")) return true;
    if (J.stringEq(s,"<=")) return true;
    if (J.stringEq(s,">=")) return true;
    return false;
  }

  boolean isTypePred(String s)
  {
    if (J.stringEq(s,"atomp")) return true;
    if (J.stringEq(s,"listp")) return true; // must have list as first element
    if (J.stringEq(s,"structurep")) return true; // includes listp 
    if (J.stringEq(s,"numberp")) return true;
    return false;
  }

  boolean isMathop(String s)
  {
    if (J.stringEq(s,"+")) return true;
    if (J.stringEq(s,"-")) return true;
    if (J.stringEq(s,"*")) return true;
    if (J.stringEq(s,"/")) return true;
    if (J.stringEq(s,"mod")) return true;
    if (J.stringEq(s,"pow")) return true;
    return false;
  }

  boolean isListOp(String s)
  {
    if (J.stringEq(s,"cons")) return true;
    if (J.stringEq(s,"first")) return true;
    if (J.stringEq(s,"rest")) return true;
    return false;
  }

  //---------------------------------------------
  // LIST OPERATIONS
  //---------------------------------------------

  // example: (cons a (list b c) (list a b c))

  JareEnv procAttCons(Vector goal,JareEnv env)
  {
    String predName=(String)goal.elementAt(0);
    Vector g2=(Vector)env.subst(goal);
    if (goal.size()!=4) die("wrong number of args: "+J.treeToString(g2));

    // shouldn't need to substitute within lists, but they could be vars
    Object a=g2.elementAt(1);
    Object b=g2.elementAt(2);
    Object c=g2.elementAt(3);

    Vector v=cons(a,b);

    if (!J.isVar(a) && isListArg(b) && J.isVar(c))
    {
      env.bind((String)c,v);
      return env;
    }
    else if (!J.isVar(a) && isListArg(b) && !J.isVar(c))
    {
      if (structEq(v,c)) return env;
      else return null;
    }
    else die("syntax of cons: arg 1 must be bound, arg 2 must be a list, and arg 3 can either be a list or var");
    return null;
  }

  boolean structEq(Object a,Object b)
  {
    if (J.isString(a) || J.isString(b))
    {
      if (!J.isString(a) || !J.isString(b)) return false;
      else return J.stringEq((String)a,(String)b);
    }
    else if (J.isVector(a) || (J.isVector(b)))
    {
      if (!J.isVector(a) || !J.isVector(b)) return false;
      Vector p=(Vector)a;
      Vector q=(Vector)b;
      if (p.size()!=q.size()) return false;
      for (int i=0 ; i<p.size() ; i++)
        if (!structEq(p.elementAt(i),q.elementAt(i))) return false;
      return true;
    }
    return false;
  }

  Vector cons(Object head,Object tail)
  {
    if (!isListArg(tail)) die("cons must be applied to list as 2nd arg");
    Vector v=(Vector)tail;
    Vector w=new Vector();
    w.addElement("list");
    w.addElement(head);
    for (int i=0 ; i<v.size()-1 ; i++) 
      w.addElement(v.elementAt(i+1));
    return w;
  }

  // handles first() and rest() - implement other list functions via rules
  // lists should not have vars or sub-structures in them
  // first arg must be list; second can be list or var

  JareEnv procAttList(Vector goal,JareEnv env)
  {
    String predName=(String)goal.elementAt(0);
    Vector g2=(Vector)env.subst(goal);
    if (goal.size()!=3) die("wrong number of args: "+J.treeToString(g2));

    Object a=g2.elementAt(1);
    Object b=g2.elementAt(2);

    if (!isListArg(a)) 
      die("list operations must have list as first arg: "+J.treeToString(g2));

    if (J.stringEq(predName,"first"))
    {
      Vector v=(Vector)a;
      if (v.size()<2) 
        die("can't take first of empty list"+J.treeToString(g2));
      if (!J.isVar(b))
      {
        if (structEq(v.elementAt(1),b)) return env;
        else return null;
      }
      else 
      {
        env.bind((String)b,v.elementAt(1));
        return env;
      }
    }
    else if (J.stringEq(predName,"rest"))
    {
      Vector v=(Vector)a;
      if (v.size()>=2) v.removeElementAt(1);
      if (!J.isVar(b))
      {
        if (structEq(v,b)) return env;
        return null;
      }
      else
      {
        env.bind((String)b,v);
        return env;
      }
    }
    return null;
  }

  boolean isListArg(Object c1)
  {
    if (!J.isVector(c1)) return false;
    Vector c2=(Vector)c1;
    if ((c2.size()>=1) && J.isString(c2.elementAt(0))
        && J.stringEq("list",(String)c2.elementAt(0))) return true;
    else return false; 
  }

  //---------------------------------------------
  // EQUALITY
  //---------------------------------------------

  // returns float or throws exception; assume already substituted

  float requireNum(Object x,String msg,Vector goal)
  {
    Float a=null;
    String errMsg="numeric argument required: ";
    if (!J.isString(x)) die(errMsg+J.treeToString(goal));
    else try { a = new Float((String)x); }
    catch (NumberFormatException e) { die(errMsg+J.treeToString(goal)); }
    return a.floatValue();
  }

  // symbol equality; both args must be instantiated
  // return env if success, null if fail

  JareEnv procAttStrEq(Vector goal,JareEnv env)
  {
    String predName=(String)goal.elementAt(0);
    Vector g2=(Vector)env.subst(goal);
    if (goal.size()!=3) die("wrong number of args: "+J.treeToString(g2));

    if (J.stringEq(predName,"eq"))
    {
      Object a1=g2.elementAt(1);
      Object b1=g2.elementAt(2);
      if (!J.isString(a1)||!J.isString(b1)) die("args to EQ can't be structured");
      String a2=(String)a1;
      String b2=(String)b1;
      if (J.isVar(a2) || J.isVar(b2)) 
        die("checking equality constraints requires bound variables: "
            +J.treeToString(g2));
      if (J.stringEq(a2,b2)) return env;
      else return null;
    }
    return null;
  }

  // handles the math equalities...
  // 2 args, must both be floats and instantiated

  JareEnv procAttMathEq(Vector goal,JareEnv env)
  {
    String predName=(String)goal.elementAt(0);
    Vector g2=(Vector)env.subst(goal);
    String errMsg="checking equality constraints requires bound variables: ";
    float a=0,b=0;

    if (goal.size()!=3) die("wrong number of args: "+J.treeToString(g2));

    Object x = g2.elementAt(1);
    a = requireNum(x,errMsg,g2);
    Object y = g2.elementAt(2);
    b = requireNum(y,errMsg,g2);

    if (J.stringEq(predName,"=") && (a==b)) return env;
    else if (J.stringEq(predName,"<") && (a<b)) return env;
    else if (J.stringEq(predName,">") && (a>b)) return env;
    else if (J.stringEq(predName,"<=") && (a<=b)) return env;
    else if (J.stringEq(predName,">=") && (a>=b)) return env;
    else return null;
  }

  // input arg already substituted    
  // 3 outcomes: 
  //   1) return a float value, 2) return null if arg is var
  //   or 3) throw exception

  Float requireNumOrVar(Object x,String msg,Vector goal)
  {
    Float a=null;
    if (!J.isString(x)) die(msg);
    else if (J.isVar((String)x)) a = null;
    else try { a = new Float((String)x); }
    catch (NumberFormatException e) { die(msg+J.treeToString(goal)); }
    return a;
  }

  //---------------------------------------------
  // MATH OPERATIONS
  //---------------------------------------------

  // for math, should work if 3 or any 2 of 3 is bound
  // if succeed, return updated environment
  // otherwise, return null
  // generally converts values to floats

  // handles: +, -, *, /, mod, pow
  // others to add: sqrt, log, exp, round
  // will solve *, -, *, / if any 2 out of 3 are bound
  // for mod, pow, etc, first 2 must be bound (3rd can be bound or not)

  JareEnv procAttMath(Vector goal,JareEnv env)
  {
    String predName=(String)goal.elementAt(0);
    Vector g2=(Vector)env.subst(goal);
    if (goal.size()!=4) die("wrong number of args: "+J.treeToString(g2));
    Float a,b,c; 
    String errMsg="math operators require numeric args: ";

    Object x=g2.elementAt(1);
    a = requireNumOrVar(x,errMsg,g2);

    Object y=g2.elementAt(2);
    b = requireNumOrVar(y,errMsg,g2);

    Object z=g2.elementAt(3);
    c = requireNumOrVar(z,errMsg,g2);
    
    if ((a!=null) && (b!=null) && (c!=null))
    {
      float af=a.floatValue();
      float bf=b.floatValue();
      float cf=c.floatValue();
      if (J.stringEq(predName,"+")&&(af+bf==cf)) return env;
      else if (J.stringEq(predName,"+")&&(af+bf==cf)) return env;
      else if (J.stringEq(predName,"-")&&(af-bf==cf)) return env;
      else if (J.stringEq(predName,"*")&&(af*bf==cf)) return env;
      else if (J.stringEq(predName,"/"))
      {
        if (bf==0) die("div by zero");
        if (af/bf==cf) return env;
      }
      else if (J.stringEq(predName,"mod")&&(af%bf==cf)) return env;
      else if (J.stringEq(predName,"pow")&&(Math.pow(af,bf)==cf)) return env;
      else return null;
    }

    if ((a==null) && (b!=null) && (c!=null))
    {
      float bf=b.floatValue();
      float cf=c.floatValue();
      if (J.stringEq(predName,"+")) env.bind((String)x,FtoS(cf-bf));
      else if (J.stringEq(predName,"-")) env.bind((String)x,FtoS(cf+bf));
      else if (J.stringEq(predName,"/")) env.bind((String)x,FtoS(cf*bf));
      else if (J.stringEq(predName,"*"))
      {
        if (bf==0) die("div by zero");
        env.bind((String)x,FtoS(cf/bf));
      }
      else if (J.stringEq(predName,"mod")) return null;
      else if (J.stringEq(predName,"pow")) return null;
      return env;
    }

    if ((a!=null) && (b==null) && (c!=null))
    {
      float af=a.floatValue();
      float cf=c.floatValue();
      if (J.stringEq(predName,"+")) env.bind((String)y,FtoS(cf-af));
      else if (J.stringEq(predName,"-")) env.bind((String)y,FtoS(af-cf));
      else if (J.stringEq(predName,"/")) 
      {
        if (cf==0) die("div by zero");
        env.bind((String)y,FtoS(af/cf));
      }
      else if (J.stringEq(predName,"*"))
      {
        if (af==0) die("div by zero");
        env.bind((String)y,FtoS(cf/af));
      }
      else if (J.stringEq(predName,"mod")) return null;
      else if (J.stringEq(predName,"pow")) return null;
      return env;
    }

    if ((a!=null) && (b!=null) && (c==null))
    {
      float af=a.floatValue();
      float bf=b.floatValue();
      if (J.stringEq(predName,"+")) env.bind((String)z,FtoS(af+bf));
      else if (J.stringEq(predName,"-")) env.bind((String)z,FtoS(af-bf));
      else if (J.stringEq(predName,"*")) env.bind((String)z,FtoS(af*bf));
      else if (J.stringEq(predName,"/"))
      {
        if (bf==0) die("div by zero");
        env.bind((String)z,FtoS(af/bf));
      }
      // mod handles negatives and floats just fine, thank you ;-)
      else if (J.stringEq(predName,"mod")) env.bind((String)z,FtoS(af%bf));
      else if (J.stringEq(predName,"pow")) 
        env.bind((String)z,new Float(Math.pow(af,bf)).toString());
      return env;
    }

    die("math operations require at least 2 of 3 args to be instantiated: "
        +J.treeToString(g2));
    return null;
  }

  String FtoS(float f) 
  { 
    return (new Float(f)).toString(); 
  }

  public void die(String s)
  {
    System.out.print("Error: ");
    System.out.println(s);
    System.exit(-1);
  }

  public void warn(String s)
  {
    System.out.println("warning: "+s);
  }
}

