// $Id: JareParser.java,v 1.2 2001/08/12 19:06:18 ioerger Exp $

package Jare;
import java.io.BufferedReader;
import java.io.FileReader;
import java.util.Stack;
import java.util.Vector;

// to do:
//   type predicates in procedural attachments
//   make testing of predicate names for proc. att. efficient via hashtables
//     i.e. hash symbols to avoid string comparisons

// done:
//   parser
//   unifer
//   back-chainer
//   next, choice-points
//   standardize variables apart
//   efficiency - at least check predicate before stdApart and unify heads
//   negation
//   procedural attachment
//   strings in parser (5/29/00)
//   use hashtables for efficiency of looking rules with matching pred names
//   tracing
//   add methods for assert and retract to engine (11/11/00)

public class JareParser
{
  public static boolean stringEq(String a,String b)
  {
    return a.equalsIgnoreCase(b);
  }

  public static boolean isString(Object x)
  {
    return (x instanceof String);
  }

  public static boolean isVector(Object x)
  {
    return (x instanceof Vector);
  }

  public static boolean isVar(Object a)
  { 
    if (!isString(a)) return false;
    String x=(String)a;
    if ((x.length()>0)&&(x.charAt(0)=='?')) return true;
    else return false;
  }

  // input nested-vector form of expression

  public static String treeToString(Object x)
  {
    if (!isVector(x)) return x.toString();
    else 
    {
      String str=new String();
      Vector v=(Vector)x;  
      str += "(";
      for (int i=0 ; i<v.size() ; i++)
      {
        if (i>0) str += " ";
        str += treeToString(v.elementAt(i));
      }
      str += ")";
      return str;
    }
  }

  // only nested lists and symbols, numbers, and strings
  // mechanism: when hit (, push current list of objects on stack & start new
  // comments are ; and // - kill to end of line

  public static Vector parse(String str) throws JareParseError
  {
    Stack lines=new Stack(); // for reporting unclosed parens
    Stack current=new Stack();
    Stack unfinished=new Stack();
    int n=str.length();
    int idx=0;
    int line=1;

    while (idx<n)
    {
      char c=str.charAt(idx);
      if (c=='(') 
      {
        unfinished.push(current);
        current = new Stack();
        lines.push(new Integer(line));
        idx++;
      }
      else if (c==')') 
      {
        if (unfinished.empty()) 
          throw new JareParseError("parse gluh error (line "+line+") - extra close paren");
        Stack temp=(Stack)unfinished.pop();
        temp.push(current);
        current = temp;
        lines.pop();
        idx++;
      }
      else if (c=='\n') { line++; idx++; }
      else if (c=='"')
      {
        int start=idx++; 
        while ((idx<n)&&(str.charAt(idx)!='"')) idx++;
        current.push(str.substring(start,idx)+'"');
        if (str.charAt(idx)=='"') idx++;
      }
      else if (c==';') 
      { 
        idx++; 
        while ((idx<n)&&(str.charAt(idx)!='\n')) idx++; 
      }
      else if ((c=='/')&&(idx<n-1))
      { 
        if (str.charAt(idx+1)=='/')
        {
          idx += 2; 
          while ((idx<n)&&(str.charAt(idx)!='\n')) idx++; 
        }
        // slightly incorrect: should NOT stop if '*/' is in quotes or in
        //   another comment like ';' or '//', but it does anyway
        else if (str.charAt(idx+1)=='*')
        {
          idx += 2; 
          while ((idx<n-1)&&((str.charAt(idx)!='*')||(str.charAt(idx+1)!='/')))
            idx++; 
          if (idx==n-1) System.out.println("warning: unclosed comment, some clauses may be skipped");
          idx += 2;
        }
      }
      else if (c=='/') // added to handle division sign
      {
        current.push(str.substring(idx,idx+1));
        idx++;
      }
      else if (!isWhitespace(c))
      {
        int start=idx;
        while ((idx<n)&&(!isWhitespace(str.charAt(idx)))
               &&(!special_char(str.charAt(idx)))
               &&(!isParen(str.charAt(idx)))) idx++;
        current.push(str.substring(start,idx));
      }
      else idx++;
    }

    if (!unfinished.empty()) 
      throw new JareParseError("parse error (line "+lines.pop()+") - missing close paren");
      //die("parse error (line "+lines.pop()+") - missing close paren");
    return (Vector)current;
  }
  
  public static boolean isParen(char c)
  {
    return ((c=='(')||(c==')'));
  }

  public static boolean isWhitespace(char c)
  {
    return ((c==' ')||(c=='\t')||(c=='\n'));
  }

  public static boolean special_char(char c)
  {
    return ((c==';')||(c=='/')||(c=='"'));
  }

  public static String read_file(String filename)
  {
    StringBuffer sb=new StringBuffer();

    try {
      String line;
      BufferedReader stream=new BufferedReader(new FileReader(filename));
      while ((line=stream.readLine())!=null) sb.append(line+'\n');
    } catch (Exception e) { System.out.println(e); }

    return new String(sb);    
  }

  public static void die(String s)
  {
    System.out.print("Error: ");
    System.out.println(s);
    System.exit(-1);
  }

  public static void warn(String s)
  {
    System.out.println("warning: "+s);
  }
}

