/*
 * $Id$
 */
package gui;

import java.awt.BorderLayout;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JTextArea;

import simulation.Environment;
import simulation.Wahrnehmung;

/**
 * @author Kai Sterker
 *
 */
public class WumpusApplikation extends JFrame {

	protected WeltDialog weltDialog = null;
	protected JButton weiter;
	private PerceptDialog perceptDialog;
	private JFrame dialog = null;
	private MapView view;
	private Environment environment = null;
	
	public static void main(String[] args) {
		WumpusApplikation w = new WumpusApplikation ();
	}
	
	public WumpusApplikation () {
		this.setSize (600, 650);
		this.setResizable (false);
		this.setTitle ("Intelligente SW-Agenten - [Die Wumpus Welt]");
		
		JPanel panel = (JPanel) this.getContentPane ();
		panel.setLayout (new BorderLayout ());
		
		view = new MapView (600, MapView.WORLD_VIEW);
		panel.add (view, BorderLayout.CENTER);

		JPanel buttons = new JPanel ();
		panel.add (buttons, BorderLayout.SOUTH);

		JButton start = new JButton ("Simulation starten");
	 	weiter = new JButton ("Naechster Schritt >>");
		final JButton welt = new JButton ("Wissen zeigen");
		
		buttons.add (start);
		buttons.add (weiter);
		buttons.add (welt);
		
		start.addActionListener (new ActionListener() {
			public void actionPerformed (ActionEvent e) {
				try {
					neueWelt ();
					weiter.setEnabled (true);
				} catch (Exception exc) { 
					exc.printStackTrace ();
				}
			}
		});

		weiter.setEnabled (false);
		weiter.addActionListener (new ActionListener() {
			public void actionPerformed (ActionEvent e) {
				try {
					if (updateSimulation () == false)
						weiter.setEnabled (false);
				} catch (Exception exc) { 
					exc.printStackTrace ();
				}
			}
		});
	
		welt.addActionListener (new ActionListener() {
			public void actionPerformed (ActionEvent e) {
				if (weltDialog == null) {
					weltDialog = new WeltDialog ();		
					weltDialog.setUmgebung (environment);
					welt.setText ("Wissen verstecken");
				}
				else {
					weltDialog.dispose ();
					weltDialog = null;
					welt.setText ("Wissen zeigen");
				}
			}
		});

		this.show ();
		
		perceptDialog = new PerceptDialog ();
	}
	
	/**
	 * 
	 */
	protected boolean updateSimulation() {
		if (environment != null) {
			int result = environment.update ();
			
			if (result == Environment.WON) {
				dialog = new ErgebnisDialog ("Der Agent hat das Gold gefunden!", "victory");
				return false;
			}
			else if (result == Environment.DEAD) {
				dialog = new ErgebnisDialog ("Der Agent ist seiner Implementierung erlegen.", "rip");
				return false;
			}
		}
		view.update (view.getGraphics());
		perceptDialog.update (environment.erzeugeWahrnehmung());
		if (weltDialog != null) weltDialog.update ();
		
		return true;
	}

	protected void neueWelt () {
		if (dialog != null) {
			dialog.dispose ();
			dialog = null;
		} 

		environment = new Environment ();
		
		if (weltDialog != null) {
			weltDialog.setUmgebung (environment);
			weltDialog.update();
		}
		
		view.setUmgebung (environment);
		view.update (view.getGraphics());
		perceptDialog.update (environment.erzeugeWahrnehmung());
	}


	private class WeltDialog extends JFrame {
		MapView welt;
		 
		public WeltDialog () {
			this.setTitle ("Wissen des Agenten");
			this.setBounds (620, 0, 400, 420);
			this.setResizable (false);
			
			JPanel panel = (JPanel) this.getContentPane ();
			panel.setLayout (new BorderLayout ());
		
			welt = new MapView (400, MapView.AGENT_VIEW);
			panel.add (welt, BorderLayout.CENTER);
			
			this.show ();
		}
		
		public void update() {
			welt.update (welt.getGraphics());
		}

		public void setUmgebung (Environment e) {
			welt.setUmgebung (e);
		}
	}

	private class ErgebnisGfx extends Canvas {
		private String gfx;
		private int offset;
		private int size;
		public ErgebnisGfx (String gfx, int offset, int size) {
			this.setBackground (Color.WHITE);
			this.setSize (size, size);
			this.gfx = gfx;
			this.offset = offset;
			this.size = size;
		}
		public void paint (Graphics g) {
			Image i = GrafikCache.getGrafik (gfx);
			g.drawImage (i, offset, offset, size, size, this);			
		}
	}
	
	private class PerceptDialog extends JFrame {
		
		public PerceptDialog () {
			this.setTitle ("Aktuelle Wahrnehmung");
			this.setBounds (620, 460, 300, 130);
			this.setResizable (false);

			JPanel panel = (JPanel) this.getContentPane ();
			panel.setLayout (new BorderLayout ());

			this.show ();
		}
		
		public void update (Wahrnehmung w) {			
			JPanel panel = (JPanel) this.getContentPane ();
			panel.removeAll();
			
			if (w.isStench ()) {
				Canvas p1 = new ErgebnisGfx ("stench", 0, 100);
				panel.add (p1, BorderLayout.WEST);
			}
			
			if (w.isBreeze ()) {
				Canvas p2 = new ErgebnisGfx ("breeze", 0, 100);
				panel.add (p2, BorderLayout.CENTER);
			}

			if (w.isGlitter ()) {
				Canvas p3 = new ErgebnisGfx ("gold", 0, 100);
				panel.add (p3, BorderLayout.EAST);
			}
		}
	}
	
	private class ErgebnisDialog extends JFrame {
		 
		public ErgebnisDialog (String message, String gfx) {
			this.setTitle ("Ergebnis");
			this.setBounds (400, 200, 200, 320);
			this.setResizable (false);
			
			JPanel panel = (JPanel) this.getContentPane ();
			panel.setLayout (new BorderLayout ());

			Canvas c = new ErgebnisGfx (gfx, 25, 150);
			c.setSize (200, 200);
			panel.add (c, BorderLayout.NORTH);
			
			Font font = new Font (null, Font.BOLD, 20);
			JTextArea m = new JTextArea ();
			m.setEditable(false);
			m.setFont(font);
			m.setLineWrap(true);
			m.setWrapStyleWord(true);
			m.setText(message);
			panel.add (m, BorderLayout.CENTER);

			JButton ok = new JButton ("Auf ein Neues!");
			panel.add (ok, BorderLayout.SOUTH);

			ok.addActionListener (new ActionListener() {
				public void actionPerformed (ActionEvent e) {
					try {
						neueWelt ();
						weiter.setEnabled (true);
					} catch (Exception exc){ }
				}
			});

			this.show ();
		}
	}
}
